# -*- coding: utf-8 -*-
# Description:
# Setup for the OSP PaloAlto Server
#
# Authors:
# Benoît Allard <benoit.allard@greenbone.net>
#
# Copyright:
# Copyright (C) 2015 Greenbone Networks GmbH
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.

import logging

from ospd.ospd import OSPDaemon, OSPDError
from ospd.misc import resolve_hostname, main as daemon_main
from ospd_paloalto import __version__

from ospd_paloalto.paloalto import PANOSClient

logger = logging.getLogger(__name__)

SEVERITIES = {
    'informational': 0.0,
    'low': 3.0,
    'medium': 5.0,
    'high': 8.0,
    'critical': 10.0,
}

OSPD_PALOALTO_DESC = """
This scanner retrieves information from a PaloAlto appliance.

There are three types of information retrieved from this
firewall solution for the specified period:

1. The logged traffic is used to find ports where a service was
   answering on the target machine.

2. The logged traffic is used to find out protocol types
   used on the ports of the target machine.

3. The logged threats are reported with the given severity
   and CVE- or other references if available.

4. For the logged threats where a CVE is referenced,
   this CVE is stored as Host Detail on attacking CVEs.

The severity classes of PaloAlto are translated into CVSS values as follows:

Informational: %s

Low: %s

Medium: %s

High: %s

Critical: %s
""" % ( SEVERITIES['informational'], SEVERITIES['low'], SEVERITIES['medium'],
        SEVERITIES['high'], SEVERITIES['critical'] )

QOD = 80

def getPorts(target, host, username, password, period, ca=None):
    client = PANOSClient(host, username ,password, ca)

    ports = {'udp': {}, 'tcp': {}}

    total = 1000
    requested = 500
    while total >= requested:
        requested *= 2
        total = 0
        logger.debug("Requesting %d Traffic entries for %s" % (requested, target))
        for traffic in client.get_traffic(target, period, requested):
            ports[traffic.protocol].setdefault(traffic.dst[1], set()).add(traffic.app[0])
            total += 1

    return ports['udp'], ports['tcp']


def getResultForPort(port, apps, period, protocol):
    s = 'Traffic analysis of %s showed replies from a service at %s port %s.' % (period, protocol.upper(), port)
    # Some entries are meaningless.
    apps -= set(['incomplete', 'insufficient-data'])
    if apps:
        s += '\n\nThe following protocol(s) have been seen on that port: %s' % (', '.join(apps))
    return dict(value=s, port="%s/%s" % (port, protocol), qod=QOD)


def gatherThreats(target, host, username, password, period, ca=None):
    client = PANOSClient(host, username ,password, ca)

    threats = {}
    vulns = {}
    
    total = 200
    requested = 100
    while total >= requested:
        requested *= 2
        total = 0
        logger.debug("Requesting %d Threats entries for %s" % (requested, target))
        for threat in client.get_threats(target, period, nresults=requested):
            if threat.id:
                vuln = client.get_vulnerability(threat.id)
                if vuln:
                    vulns[threat.id] = vuln
            if threat in threats:
                threats[threat].count += threat.count
            else:
                threats[threat] = threat
            total += 1

    return threats.values(), vulns


def getResultForThreat(threat, vuln, period):
    s = ["Threat analysis of %s showed that the threat has been experienced %d times." % (period, threat.count)]
    if vuln:
        s.append("\nThis vulnerability has a severity of %s." % (vuln.severity))
        if vuln.cves:
            s.append("The following CVE(s) are associated with this vulnerability: %s." % ', '.join(vuln.cves))
        if vuln.vendors:
            s.append("This vulnerability is also known as %s." % ', '.join(vuln.vendors))
    s.append('This Vulnerability is known under the PaloAlto identifier %d.' % threat.id)
    s.append('\nThe following action has been undertaken: "%s"' % threat.action)
    return dict(
        name=threat.name,
        value='\n'.join(s),
        port="%s/%s" % (threat.destination[1], threat.protocol),
        severity=str(SEVERITIES[threat.severity]),
        qod=QOD)

WRAPPER_SETTINGS = {
    'address': {
        'type': 'string',
        'name': 'PaloAlto address',
        'default': '',
        'description': 'The hostname or IP where to reach the PaloAlto appliance.',
    },
    'username': {
        'type': 'string',
        'name': 'Login name',
        'default': '',
        'description': 'The name of the account used to login to the PaloAlto appliance.',
    },
    'password': {
        'type': 'password',
        'name': 'Password credential',
        'default': '',
        'description': 'The password used to login to the PaloAlto appliance.',
    },
    'period': {
        'type': 'selection',
        'name': 'Time period to analyze',
        'default': 'last-6-hrs|last-12-hrs|last-24-hrs|last-7-days|last-30-days',
        'description': 'The traffic for the specified period will be analysed.',
    },
    'ca-certificate': {
        'type': 'file',
        'name': 'CA Certificate',
        'default': '',
        'description': 'CA certificate used to issue the PaloAlto HTTPS certificate',
    }
}

class OSPDWrapper(OSPDaemon):

    """ Class for ospd-paloalto daemon. """

    def __init__(self, certfile, keyfile, cafile):
        """ Initializes the ospd-paloalto daemon's internal data. """
        super(OSPDWrapper, self).__init__(certfile=certfile, keyfile=keyfile,
            cafile=cafile)
        self.server_version = __version__
        self.scanner_info['name'] = 'PaloAlto'
        self.scanner_info['version'] = 'depends on the version of the appliance'
        self.scanner_info['description'] = OSPD_PALOALTO_DESC
        for name, params in WRAPPER_SETTINGS.items():
            self.add_scanner_param(name, params)

    def process_scan_params(self, params):
        """ params is directly from the XML """
        for param in WRAPPER_SETTINGS:
            if param in ('ca-certificate',):
                continue
            if not params.get(param):
                raise OSPDError('Empty %s value' % param, 'start_scan')
        return params

    def check(self):
        """ Checks that the PaloAlto system is found. """
        return True

    def exec_scan(self, scan_id, target):
        self.getPorts(scan_id, target)
        self.getThreats(scan_id, target)

    def getPorts(self, scan_id, target):
        options = self.get_scan_options(scan_id)

        udp, tcp = getPorts(target, options['address'], options['username'], options['password'], options['period'], options['ca-certificate'])

        target_ip = resolve_hostname(target)
        if target_ip is None:
            logger.info("Couldn't resolve {0}.".format(target))
            target_ip = target
        # We need at least one result to get the host details registered.
        self.add_scan_log(scan_id, host=target_ip, name='PaloAlto open port summary',
            value='In the %s, services at %d ports answered on the target host (%d TCP and %d UDP).' % (options['period'], len(tcp) + len(udp), len(tcp), len(udp)), qod=QOD)

        # Create a log entry for each found port
        for port in udp:
            self.add_scan_log(scan_id, host=target_ip, name='PaloAlto open port detection',
                              **getResultForPort(port, udp[port], options['period'], 'udp'))
        for port in tcp:
            self.add_scan_log(scan_id, host=target_ip, name='PaloAlto open port detection',
                              **getResultForPort(port, tcp[port], options['period'], 'tcp'))

        if tcp or udp:
            self.add_scan_host_detail(scan_id, host=target_ip, name="ports",
                                      value=", ".join(str(p) for p in sorted(list(tcp) + list(udp))))
        if udp:
            self.add_scan_host_detail(scan_id, host=target_ip, name="udp_ports",
                                      value=", ".join(str(p) for p in sorted(udp)))
        if tcp:
            self.add_scan_host_detail(scan_id, host=target_ip, name="tcp_ports",
                                      value=", ".join(str(p) for p in sorted(tcp)))

    def getThreats(self, scan_id, target):
        options = self.get_scan_options(scan_id)

        threats, vulns = gatherThreats(target, options['address'], options['username'], options['password'], options['period'], options['ca-certificate'])

        total = sum(threat.count for threat in threats)

        self.add_scan_log(scan_id, host=target, name="PaloAlto threats summary",
            value="In the %s, %s threats have been logged, with a total of %d distinct Vulnerabilities." % (options['period'], total, len(vulns)), qod=QOD)

        CVEs = set()
        for threat in threats:
            vuln = vulns.get(threat.id)
            if vuln:
                CVEs.update(vuln.cves)
            self.add_scan_alarm(scan_id, host=target,
                **getResultForThreat(threat, vuln, options['period']))

        if CVEs:
            self.add_scan_host_detail(scan_id, host=target, name="cves_attacks", value=", ".join(sorted(CVEs)))

def main():
    """ OSP PaloAlto main function. """
    daemon_main('OSPD - PaloAlto wrapper', OSPDWrapper)
